//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
using System;
using System.Collections.Generic;

namespace Nocandy.LinqResources.Mystery
{
  /// <summary>
  /// Success, Failure, or "Unknown error".
  /// </summary>
  public interface Mystery<E, out A>
  {
    /// <summary>
    /// Catamorphism.
    /// </summary>
    /// <param name='unknown'>
    /// Unknown case.
    /// </param>
    /// <param name='failure'>
    /// Failure case.
    /// </param>
    /// <param name='success'>
    /// Success case.
    /// </param>
    /// <typeparam name='Z'>
    /// The result type.
    /// </typeparam>
    Z Fold<Z>(Func<string, Z> unknown, Func<E, Z> failure, Func<A, Z> success);

    #region Functor
    /// <summary>
    /// Modify the result of my success.
    /// </summary>
    /// <param name='f'>
    /// Lifted function.
    /// </param>
    /// <typeparam name='B'>
    /// The new result type.
    /// </typeparam>
    Mystery<E, B> Select<B>(Func<A, B> f);

    /// <summary>
    /// Modify any given error.
    /// </summary>
    /// <returns>
    /// Myself with error modified, if applicable.
    /// </returns>
    /// <param name='errorf'>
    /// The error homomorphism.
    /// </param>
    /// <typeparam name='E2'>
    /// The new error type.
    /// </typeparam>
    Mystery<E2, A> SelectError<E2>(Func<E, E2> errorf);
    #endregion

    #region Bind
    /// <summary>
    /// Monadic bind.
    /// </summary>
    /// <returns>
    /// Running myself through f, then final.
    /// </returns>
    /// <param name='f'>
    /// Produce actions from my result.
    /// </param>
    /// <param name='final'>
    /// What to do with my result and that of f.
    /// </param>
    Mystery<E, Z> SelectMany<B, Z>(Func<A, Mystery<E, B>> f, Func<A, B, Z> final);
    #endregion

    #region Plus
    /// <summary>
    /// Filter with pred.
    /// </summary>
    /// <param name='pred'>
    /// Fail if it doesn't pass.
    /// </param>
    Mystery<E, A> Where(Func<A, bool> pred);

    /// <summary>
    /// Join on keys, running myself first.
    /// </summary>
    /// <param name='bs'>
    /// Second action.
    /// </param>
    /// <param name='ak'>
    /// Left side of the filter.
    /// </param>
    /// <param name='bk'>
    /// Right side of the filter.
    /// </param>
    /// <param name='final'>
    /// Final mapped result.
    /// </param>
    Mystery<E, Z> Join<B, K, Z>(Mystery<E, B> bs, Func<A, K> ak, Func<B, K> bk, Func<A, B, Z> final);
    #endregion
  }

  public abstract class AMystery<E, A>: Mystery<E, A>
  {
    public abstract Z Fold<Z>(Func<string, Z> unknown, Func<E, Z> failure, Func<A, Z> success);
    public abstract Mystery<E, B> Select<B>(Func<A, B> f);
    public abstract Mystery<E2, A> SelectError<E2>(Func<E, E2> errorf);
    public abstract Mystery<E, Z> SelectMany<B, Z>(Func<A, Mystery<E, B>> f, Func<A, B, Z> final);
    public abstract Mystery<E, A> Where(Func<A, bool> pred);

    #region Bind
    /// <summary>
    /// Sequence outer, then inner action.
    /// </summary>
    /// <param name='me'>
    /// A mystery to flatten.
    /// </param>
    public static Mystery<E, A> Join(Mystery<E, Mystery<E, A>> me)
    {
      return me.SelectMany<A, A>(mea => mea, (ign, a) => a);
    }
    #endregion

    #region Pointed
    /// <summary>
    /// Value without action.
    /// </summary>
    /// <param name='val'>
    /// The success.
    /// </param>
    public static Mystery<E, A> Point(A val)
    {
      return new Success<E, A>(val);
    }
    #endregion

    #region Plus
    public Mystery<E, Z> Join<B, K, Z>(Mystery<E, B> bs, Func<A, K> ak, Func<B, K> bk, Func<A, B, Z> final)
    {
      return from anA in this from aB in bs
             where ak(anA).Equals(bk(aB))
             select final(anA, aB);
    }

    /// <summary>
    /// Monoidal identity.
    /// </summary>
    /// <returns>
    /// Skipped in Plus, short-circuits Join.
    /// </returns>
    public static Mystery<E, A> Zero()
    {
      return new UnknownError<E, A>("short-circuit");
    }

    /// <summary>
    /// Semigroup append.
    /// </summary>
    /// <returns>
    /// Myself if I succeed, right otherwise.
    /// </returns>
    /// <param name='right'>
    /// The action to take if I fail.
    /// </param>
    public static Mystery<E, A> Plus(Mystery<E, A> left,
                                     Mystery<E, A> right)
    {
      return left.Fold (s => right, e => right, a => left);
    }
    #endregion

    #region structural equality
    public override bool Equals (object other)
    {
      if (base.Equals (other)) return true;
      var o = other as Mystery<E, A>;
      if (ReferenceEquals (null, o)) return false;
      return Fold
        (a => o.Fold (b => a == b, z => false, z => false),
         a => o.Fold (z => false, b => Equals (a, b), z => false),
         a => o.Fold (z => false, z => false, b => Equals (a, b)));
    }

    public override int GetHashCode ()
    {
      return Fold
        (a => new Tuple<int, string>(4, a).GetHashCode (),
         a => new Tuple<int, E>(8, a).GetHashCode (),
         a => new Tuple<int, A>(15, a).GetHashCode ());
    }
    #endregion
  }

  public static class MysteryOps
  {
    #region Applicative (Mystery e)
    /// <summary>
    /// Run a bunch of Mysteries.
    /// </summary>
    /// <returns>
    /// Success only if all succeeded, the first failure otherwise.
    /// </returns>
    public static Mystery<E, IEnumerable<A>> Sequence<E, A>
      (this IEnumerable<Mystery<E, A>> maybes)
    {
      var acc = new List<A>();
      Mystery<E, IEnumerable<A>> cap = null;
      foreach (var elt in maybes)
      {
        elt.Fold (delegate(String m)
                  {
                    cap = new UnknownError<E, IEnumerable<A>>(m);
                    return 0;
                  },
                  delegate(E e)
                  {
                    cap = new Failure<E, IEnumerable<A>>(e);
                    return 0;
                  },
                  delegate(A a){acc.Add(a); return 0;});
        if (null != cap) return cap;
      }
      return AMystery<E, IEnumerable<A>>.Point (acc);
    }
    #endregion
  }

  public class Success<E, A>: AMystery<E, A>
  {
    private A Result {get; set;}

    public Success(A val)
    {
      Result = val;
    }

    public override Mystery<E, B> Select<B>(Func<A, B> f)
    {
      return new Success<E, B>(f(Result));
    }

    public override Mystery<E2, A> SelectError<E2>(Func<E, E2> errorf)
    {
      return new Success<E2, A>(Result);
    }

    public override Mystery<E, Z> SelectMany<B, Z>(Func<A, Mystery<E, B>> f, Func<A, B, Z> final)
    {
      return from b in f(Result) select final(Result, b);
    }

    public override Mystery<E, A> Where(Func<A, bool> pred)
    {
      if (pred(Result))
        return this;
      else
        return new UnknownError<E, A>("filtered out");
    }

    public override Z Fold<Z>(Func<string, Z> unknown, Func<E, Z> failure, Func<A, Z> success)
    {
      return success(Result);
    }

    public override string ToString ()
    {
      return string.Format ("Success({0})", Result);
    }
  }

  public class Failure<E, A>: AMystery<E, A>
  {
    private E Error {get; set;}

    public Failure(E err)
    {
      Error = err;
    }

    public Mystery<E, B> AnySuccess<B>()
    {
      return new Failure<E, B>(Error);
    }

    public override Mystery<E, B> Select<B>(Func<A, B> f)
    {
      return AnySuccess<B>();
    }

    public override Mystery<E2, A> SelectError<E2>(Func<E, E2> errorf)
    {
      return new Failure<E2, A>(errorf(Error));
    }

    public override Mystery<E, Z> SelectMany<B, Z>(Func<A, Mystery<E, B>> f, Func<A, B, Z> final)
    {
      return AnySuccess<Z>();
    }

    public override Mystery<E, A> Where(Func<A, bool> pred)
    {
      return this;
    }

    public override Z Fold<Z>(Func<string, Z> unknown, Func<E, Z> failure, Func<A, Z> success)
    {
      return failure(Error);
    }

    public override string ToString ()
    {
      return string.Format ("Failure({0})", Error);
    }
  }

  public class UnknownError<E, A>: AMystery<E, A>
  {
    private string Message {get; set;}

    public UnknownError() : this("unknown error") {}

    public UnknownError(string message)
    {
      Message = message;
    }

    public Mystery<E, B> AnySuccess<B>()
    {
      return new UnknownError<E, B>(Message);
    }

    public override Mystery<E, B> Select<B>(Func<A, B> f)
    {
      return AnySuccess<B>();
    }

    public override Mystery<E2, A> SelectError<E2>(Func<E, E2> errorf)
    {
      return new UnknownError<E2, A>(Message);
    }

    public override Mystery<E, Z> SelectMany<B, Z>(Func<A, Mystery<E, B>> f, Func<A, B, Z> final)
    {
      return AnySuccess<Z>();
    }

    public override Mystery<E, A> Where(Func<A, bool> pred)
    {
      return this;
    }

    public override Z Fold<Z>(Func<string, Z> unknown, Func<E, Z> failure, Func<A, Z> success)
    {
      return unknown(Message);
    }

    public override string ToString ()
    {
      return string.Format ("UnknownError({0})", Message);
    }
  }
}

