//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

using System;
using System.Collections.Generic;
using System.Linq;

namespace Nocandy.LinqResources.Reader
{
  /// <summary>
  /// The requirement that, to produce a TResult, you must have a 
  /// TInput.
  /// </summary>
  public delegate TResult Reader<in TInput, out TResult>(TInput requirement);

  /// <summary>
  /// Extension methods on Reader, and reader constructors.
  /// </summary>
  public static class ReaderOps
  {
    #region Contravariant
    /// <summary>
    /// Reader forms a contravariant functor on A.
    /// </summary>
    /// <param name='g'>
    /// Extract an A.
    /// </param>
    /// <typeparam name='B'>
    /// The new required type.
    /// </typeparam>
    public static Reader<B, R> Contramap<A, R, B>(this Reader<A, R> f, Func<B, A> g)
    {
      return Arrow(g).Compose(f);
    }
    #endregion

    #region Category
    /// <summary>
    /// Morphism composition.
    /// </summary>
    /// <param name='andThen'>
    /// What to do with my result.
    /// </param>
    /// <typeparam name='S'>
    /// The implied result of the composition.
    /// </typeparam>
    public static Reader<A, S> Compose<A, R, S>(this Reader<A, R> f, Reader<R, S> andThen)
    {
      return a => andThen(f(a));
    }

    /// <summary>
    /// Categorical identity.
    /// </summary>
    /// <typeparam name='I'>
    /// Universally quantified.
    /// </typeparam>
    public static Reader<I, I> Identity<I>()
    {
      return x => x;
    }

    /// <summary>
    /// Alias for Identity.
    /// </summary>
    public static Reader<I, I> Get<I>()
    {
      return Identity<I> ();
    }
    #endregion

    #region Arrow
    /// <summary>
    /// Lift a function.
    /// </summary>
    /// <param name='f'>
    /// Ordinary function.
    /// </param>
    public static Reader<A, R> Arrow<A, R>(Func<A, R> f)
    {
      return x => f(x);
    }

    /// <summary>
    /// Pass through data.
    /// </summary>
    /// <typeparam name='S'>
    /// Data to pass through.
    /// </typeparam>
    public static Reader<Tuple<A, S>, Tuple<R, S>> First<A, R, S>(this Reader<A, R> f)
    {
      return tup => new Tuple<R, S>(f(tup.Item1), tup.Item2);
    }

    /// <summary>
    /// Two readers can run side-by-side.
    /// </summary>
    /// <param name='fst'>
    /// Item1 morphism.
    /// </param>
    /// <param name='snd'>
    /// Item2 morphism.
    /// </param>
    public static Reader<Tuple<A, A2>, Tuple<R, R2>> Product<A, R, A2, R2>(this Reader<A, R> fst, Reader<A2, R2> snd)
    {
      return tup => new Tuple<R, R2>(fst(tup.Item1), snd(tup.Item2));
    }

    /// <summary>
    /// Send input to two readers.
    /// </summary>
    /// <param name='fst'>
    /// Produce Item1.
    /// </param>
    /// <param name='snd'>
    /// Produce Item2.
    /// </param>
    public static Reader<A, Tuple<R, S>> Fanout<A, R, S>(this Reader<A, R> fst, Reader<A, S> snd)
    {
      return a => new Tuple<R, S>(fst(a), snd(a));
    }
    #endregion

    #region Functor
    /// <summary>
    /// Map the result to something else.
    /// </summary>
    /// <param name='f'>
    /// The reader context.
    /// </param>
    /// <param name='then'>
    /// What to do with the result.
    /// </param>
    public static Reader<A, S> Select<A, R, S>(this Reader<A, R> f, Func<R, S> then)
    {
      return a => then(f(a));
    }
    #endregion

    #region Distributive
    /// <summary>
    /// Provide an input to a whole sequence of readers.
    /// </summary>
    public static Reader<A, IEnumerable<R>> Distribute<A, R>
      (this IEnumerable<Reader<A, R>> fs)
    {
      return a => from f in fs select f(a);
    }
    #endregion

    #region Pointed
    /// <summary>
    /// A pure value.
    /// </summary>
    /// <param name='val'>
    /// Value to use, always.
    /// </param>
    public static Reader<A, R> Point<A, R>(R val)
    {
      return t => val;
    }
    #endregion

    #region Apply
    /// <summary>
    /// Map two results to something else.
    /// </summary>
    /// <param name='l'>
    /// Do this first.
    /// </param>
    /// <param name='r'>
    /// Do this second.
    /// </param>
    /// <param name='f'>
    /// Do this last.
    /// </param>
    public static Reader<A, Z> Select2<A, R, S, Z>(this Reader<A, R> l, Reader<A, S> r, Func<R, S, Z> f)
    {
      return anA => f(l(anA), r(anA));
    }

    /// <summary>
    /// Run effect, then myself.
    /// </summary>
    /// <returns>
    /// The receiver's value.
    /// </returns>
    public static Reader<A, R> ButFirst<A, R, S>(this Reader<A, R> real, Reader<A, S> effect)
    {
      return effect.Select2(real, (ign, it) => it);
    }

    /// <summary>
    /// Run myself, then effect.
    /// </summary>
    /// <returns>
    /// The receiver's value.
    /// </returns>
    public static Reader<A, R> ButLast<A, R, S>(this Reader<A, R> real, Reader<A, S> effect)
    {
      return real.Select2(effect, (r, ign) => r);
    }
    #endregion

    #region Bind
    /// <summary>
    /// Monadic bind.
    /// </summary>
    /// <returns>
    /// Some Z.
    /// </returns>
    /// <param name='xs'>
    /// Starting value.
    /// </param>
    /// <param name='bind'>
    /// Produce an additional value.
    /// </param>
    /// <param name='end'>
    /// Map those values to something else.
    /// </param>
    public static Reader<A, Z> SelectMany<A, R, S, Z>(this Reader<A, R> xs, Func<R, Reader<A, S>> bind, Func<R, S, Z> end)
    {
      return delegate(A a)
      {
        var anR = xs(a);
        return end(anR, bind(anR)(a));
      };
    }

    /// <summary>
    /// Sequence outer, then inner action.
    /// </summary>
    /// <param name='ff'>
    /// Action to sequence.
    /// </param>
    public static Reader<A, R> Join<A, R>(this Reader<A, Reader<A, R>> ff)
    {
      return anA => ff(anA)(anA);
    }
    #endregion
  }
}

