//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.

using System;
using System.Collections.Generic;
using System.Linq;
using Nocandy.LinqResources.Mystery;

namespace Nocandy.LinqResources.ReaderV
{
  /// <summary>
  /// The requirement that, to produce a TResult, you must have a 
  /// TInput, and not an Exception.
  /// </summary>
  /// <remarks>
  ///   <para>
  ///     I describe operations that are delayed until a particular
  ///     piece of data, TInput, is given.  Once TInput is provided,
  ///     the operation can still fail, which is encoded as a Mystery.
  ///   </para>
  ///   <para>
  ///     ReaderVOps provides many LINQ operators, and a few others,
  ///     for creating, combining, and manipulating ReaderVs.
  ///   </para>
  ///   <para>
  ///     I recommend that, despite the convenience of delegate syntax
  ///     in C#, you avoid temptation and build primitives from the
  ///     Get and Lift methods.
  ///   </para>
  /// </remarks>
  public delegate Mystery<Exception, TResult>
    ReaderV<in TInput, out TResult>(TInput requirement);

  /// <summary>
  /// Extension methods on ReaderV, and reader constructors.
  /// </summary>
  public static class ReaderVOps
  {
    #region Contravariant
    /// <summary>
    /// ReaderV forms a contravariant functor on A.
    /// </summary>
    /// <param name='g'>
    /// Extract an A.
    /// </param>
    /// <typeparam name='B'>
    /// The new required type.
    /// </typeparam>
    public static ReaderV<B, R> Contramap<A, R, B>(this ReaderV<A, R> f, Func<B, A> g)
    {
      return Arrow(g).Compose(f);
    }
    #endregion

    #region Category
    /// <summary>
    /// Morphism composition.
    /// </summary>
    /// <param name='andThen'>
    /// What to do with my result.
    /// </param>
    /// <typeparam name='S'>
    /// The implied result of the composition.
    /// </typeparam>
    public static ReaderV<A, S> Compose<A, R, S>(this ReaderV<A, R> f, ReaderV<R, S> andThen)
    {
      return a => from r in f(a) from s in andThen(r) select s;
    }

    /// <summary>
    /// Categorical identity.
    /// </summary>
    /// <typeparam name='I'>
    /// Universally quantified.
    /// </typeparam>
    public static ReaderV<I, I> Identity<I>()
    {
      return x => AMystery<Exception, I>.Point (x);
    }

    /// <summary>
    /// Alias for Identity.
    /// </summary>
    public static ReaderV<I, I> Get<I>()
    {
      return Identity<I> ();
    }
    #endregion

    #region Arrow
    /// <summary>
    /// Lift a function.
    /// </summary>
    /// <param name='f'>
    /// Ordinary function.
    /// </param>
    public static ReaderV<A, R> Arrow<A, R>(Func<A, R> f)
    {
      return x => AMystery<Exception, R>.Point (f(x));
    }

    /// <summary>
    /// Pass through data.
    /// </summary>
    /// <typeparam name='S'>
    /// Data to pass through.
    /// </typeparam>
    public static ReaderV<Tuple<A, S>, Tuple<R, S>> First<A, R, S>(this ReaderV<A, R> f)
    {
      return tup => from r in f(tup.Item1)
        select new Tuple<R, S>(r, tup.Item2);
    }

    /// <summary>
    /// Two readers can run side-by-side.
    /// </summary>
    /// <param name='fst'>
    /// Item1 morphism.
    /// </param>
    /// <param name='snd'>
    /// Item2 morphism.
    /// </param>
    public static ReaderV<Tuple<A, A2>, Tuple<R, R2>> Product<A, R, A2, R2>(this ReaderV<A, R> fst, ReaderV<A2, R2> snd)
    {
      return tup => from l in fst(tup.Item1) from r in snd(tup.Item2)
        select new Tuple<R, R2>(l, r);
    }

    /// <summary>
    /// Send input to two readers.
    /// </summary>
    /// <param name='fst'>
    /// Produce Item1.
    /// </param>
    /// <param name='snd'>
    /// Produce Item2.
    /// </param>
    public static ReaderV<A, Tuple<R, S>> Fanout<A, R, S>(this ReaderV<A, R> fst, ReaderV<A, S> snd)
    {
      return a => from l in fst(a) from r in snd(a)
        select new Tuple<R, S>(l, r);
    }
    #endregion

    #region Functor
    /// <summary>
    /// Map the result to something else.
    /// </summary>
    /// <param name='f'>
    /// The reader context.
    /// </param>
    /// <param name='then'>
    /// What to do with the result.
    /// </param>
    public static ReaderV<A, S> Select<A, R, S>(this ReaderV<A, R> f, Func<R, S> then)
    {
      return a => from r in f(a) select then(r);
    }
    #endregion

    #region Distributive
    /// <summary>
    /// Provide an input to a whole sequence of readers.
    /// </summary>
    public static ReaderV<A, IEnumerable<R>> Distribute<A, R>
      (this IEnumerable<ReaderV<A, R>> fs)
    {
      return a => (from f in fs select f(a)).Sequence();
    }
    #endregion

    #region Pointed
    /// <summary>
    /// A pure value.
    /// </summary>
    /// <param name='val'>
    /// Value to use, always.
    /// </param>
    public static ReaderV<A, R> Point<A, R>(R val)
    {
      return t => AMystery<Exception, R>.Point (val);
    }
    #endregion

    #region Apply
    /// <summary>
    /// Map two results to something else.
    /// </summary>
    /// <param name='l'>
    /// Do this first.
    /// </param>
    /// <param name='r'>
    /// Do this second.
    /// </param>
    /// <param name='f'>
    /// Do this last.
    /// </param>
    public static ReaderV<A, Z> Select2<A, R, S, Z>(this ReaderV<A, R> l, ReaderV<A, S> r, Func<R, S, Z> f)
    {
      return anA => from ro in l(anA) from so in r(anA) select f(ro, so);
    }

    /// <summary>
    /// Run effect, then myself.
    /// </summary>
    /// <returns>
    /// The receiver's value.
    /// </returns>
    public static ReaderV<A, R> ButFirst<A, R, S>(this ReaderV<A, R> real, ReaderV<A, S> effect)
    {
      return effect.Select2(real, (ign, it) => it);
    }

    /// <summary>
    /// Run myself, then effect.
    /// </summary>
    /// <returns>
    /// The receiver's value.
    /// </returns>
    public static ReaderV<A, R> ButLast<A, R, S>(this ReaderV<A, R> real, ReaderV<A, S> effect)
    {
      return real.Select2(effect, (r, ign) => r);
    }
    #endregion

    #region Bind
    /// <summary>
    /// Monadic bind.
    /// </summary>
    /// <returns>
    /// Some Z.
    /// </returns>
    /// <param name='xs'>
    /// Starting value.
    /// </param>
    /// <param name='bind'>
    /// Produce an additional value.
    /// </param>
    /// <param name='end'>
    /// Map those values to something else.
    /// </param>
    public static ReaderV<A, Z> SelectMany<A, R, S, Z>(this ReaderV<A, R> xs, Func<R, ReaderV<A, S>> bind, Func<R, S, Z> end)
    {
      return a => from anR in xs(a) from anS in bind(anR)(a)
        select end(anR, anS);
    }

    /// <summary>
    /// Sequence outer, then inner action.
    /// </summary>
    /// <param name='ff'>
    /// Action to sequence.
    /// </param>
    public static ReaderV<A, R> Join<A, R>(this ReaderV<A, ReaderV<A, R>> ff)
    {
      return anA => AMystery<Exception, R>.Join(from rr in ff(anA) select rr(anA));
    }
    #endregion

    #region Plus
    /// <summary>
    /// Monoidal identity.
    /// </summary>
    /// <returns>
    /// A default short-circuit.  Note that any unsuccessful mystery
    /// result is a short-circuit for Plus, SelectMany, Join, and any
    /// other operation requiring an R, as there are no Rs in
    /// unsuccessful mysteries.
    /// </returns>
    public static ReaderV<A, R> Zero<A, R>()
    {
      return a => AMystery<Exception, R>.Zero();
    }

    /// <summary>
    /// Semigroup append.
    /// </summary>
    /// <returns>
    /// Myself if I succeed, right otherwise.
    /// </returns>
    /// <param name='right'>
    /// The action to take if I fail.
    /// </param>
    public static ReaderV<A, R> Plus<A, R>(this ReaderV<A, R> left,
                                           ReaderV<A, R> right)
    {
      return delegate(A a)
      {
        var lr = left(a);
        return lr.Fold(s => right(a), e => right(a), r => lr);
      };
    }

    /// <summary>
    /// Filter with pred.
    /// </summary>
    /// <param name='pred'>
    /// Fail if it doesn't pass.
    /// </param>
    public static ReaderV<A, R> Where<A, R>
      (this ReaderV<A, R> self, Func<R, bool> pred)
    {
      return a => self(a).Where(pred);
    }

    /// <summary>
    /// Join on keys, running myself first.
    /// </summary>
    /// <param name='bs'>
    /// Second action.
    /// </param>
    /// <param name='ak'>
    /// Left side of the filter.
    /// </param>
    /// <param name='bk'>
    /// Right side of the filter.
    /// </param>
    /// <param name='final'>
    /// Final mapped result.
    /// </param>
    public static ReaderV<A, Z> Join<A, R, B, K, Z>
      (this ReaderV<A, R> rs, ReaderV<A, B> bs,
       Func<R, K> ak, Func<B, K> bk, Func<R, B, Z> final)
    {
      return from r in rs from b in bs where Equals(ak(r), bk(b))
        select final(r, b);
    }
    #endregion

    #region of Mysteries
    public static Mystery<Exception, R> Run<A, R>
      (this ReaderV<A, R> self, A input)
    {
      return self(input);
    }
    #endregion
  }

  /// <summary>
  ///   Extensions for Mysteries.
  /// </summary>
  public static class ReaderVMysteryOps
  {
    #region of Mysteries
    /// <summary>
    ///   Ignore A, producing result.
    /// </summary>
    /// <param name='result'>
    ///   The success, failure, or what have you that should always result.
    /// </param>
    /// <returns>
    ///   A ReaderV that ignores its input, answering result.
    /// </returns>
    public static ReaderV<A, R> Lift<A, R>(this Mystery<Exception, R> result)
    {
      return a => result;
    }
    #endregion
  }

  /// <summary>
  ///   Extensions for universal values.
  /// </summary>
  public static class ReaderVAnyOps
  {
    /// <summary>
    ///   Succeed.
    /// </summary>
    public static ReaderV<A, R> RVSuccess<A, R>(this R self)
    {
      return ReaderVOps.Point<A, R>(self);
    }

    /// <summary>
    ///   Fail with an exception explaining the problem.
    /// </summary>
    public static ReaderV<A, R> RVFailure<A, R>(this Exception self)
    {
      return new Failure<Exception, R>(self).Lift<A, R>();
    }

    /// <summary>
    ///   Fail with a string describing the problem.
    /// </summary>
    public static ReaderV<A, R> RVUnknownError<A, R>(this string self)
    {
      return new UnknownError<Exception, R>(self).Lift<A, R>();
    }
  }
}

