//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
using System;
using System.Collections.Generic;
using Nocandy.LinqResources.ReaderV;

namespace Nocandy.LinqResources.ReaderVExample
{
  #region Program data, the contextual information about a program
  /// <summary>
  ///   Information about the currently running program.
  /// </summary>
  public class Program
  {
    /// <summary>
    ///   The current user.
    /// </summary>
    public UserInfo User {get; private set;}

    /// <summary>
    ///   String settings from configuration.
    /// </summary>
    public Func<string, string> Settings {get; private set;}

    /// <summary>
    ///   Create.
    /// </summary>
    /// <param name='user'>
    ///   Value of User property.
    /// </param>
    /// <param name='settings'>
    ///   Value of Settings property.
    /// </param>
    public Program(UserInfo user, Func<string, string> settings)
    {
      User = user;
      Settings = settings;
    }
  }

  /// <summary>
  ///   Information about a user.
  /// </summary>
  public class UserInfo
  {
    /// <summary>
    ///   The user's name.
    /// </summary>
    public string Name {get; private set;}

    /// <summary>
    ///   The user's age.
    /// </summary>
    public int Age {get; private set;}

    /// <summary>
    ///   Create.
    /// </summary>
    /// <param name='name'>
    ///   The Name property.
    /// </param>
    /// <param name='age'>
    ///   The Age property.
    /// </param>
    public UserInfo(string name, int age)
    {
      Name = name;
      Age = age;
    }
  }
  #endregion Program data, the contextual information about a program

  #region Trying out these toys
  public class Scaffold
  {
    /// <summary>
    ///   Run the "program".
    /// </summary>
    /// <param name='name'>
    ///   The user's name.
    /// </param>
    /// <param name='age'>
    ///   The user's age.
    /// </param>
    /// <param name='settings'>
    ///   Program configuration properties.
    /// </param>
    /// <returns>
    ///   A message about the user, or an error message explaining
    ///   what went wrong.
    /// </returns>
    public static string Run(string name, int age,
                             IDictionary<string, string> settings)
    {
      return (Describe.ButFirst(MustBeRunnable2))
        .Run(new Program(user: new UserInfo(name: name, age: age),
                         settings: settings.GetFunc ()))
        .Fold(s => string.Format ("ERROR: {0}", s),
              e => string.Format ("Exception: {0}", e.Message),
              a => a);
    }

    /// <summary>
    ///   Succeed iff the program can be run.
    /// </summary>
    public static ReaderV<Program, Unit> MustBeRunnable()
    {
      return from prg in ReaderVOps.Get<Program>()
        where prg.Settings("runnable") == "yes"
        select Unit.Instance;
    }

    /// <summary>
    ///   Succeed iff the program can be run.  Like MustBeRunnable,
    ///   but with a better error.
    /// </summary>
    public static readonly ReaderV<Program, Unit> MustBeRunnable2 =
      from prg in ReaderVOps.Get<Program>()
      from u in (prg.Settings("runnable") == "yes"
                 ? Unit.Instance.RVSuccess<Program, Unit>()
                 : "you must set runnable=yes".RVUnknownError<Program, Unit>())
      select u;

    /// <summary>
    ///   Whether the program should be run verbosely.
    /// </summary>
    public static readonly ReaderV<Program, bool> Verbose =
      from prg in ReaderVOps.Get<Program>()
      select prg.Settings ("verbose") == "yes";

    /// <summary>
    ///   Describe the user, however the configuration says to.
    /// </summary>
    public static readonly ReaderV<Program, string> Describe =
      from vb in Verbose
      from msg in (DescribeLong().Where(s => vb)
                   .Plus(DescribeShort())
                   .Contramap((Program p) => p.User))
      select msg;

    /// <summary>
    ///   Describe the user in a long way.
    /// </summary>
    public static ReaderV<UserInfo, string> DescribeLong ()
    {
      return from ui in ReaderVOps.Get<UserInfo>()
        select string.Format ("The user {0} is of age {1}",
                              ui.Name, ui.Age);
    }

    /// <summary>
    ///   Describe the user in a short way.
    /// </summary>
    public static ReaderV<UserInfo, string> DescribeShort ()
    {
      return from ui in ReaderVOps.Get<UserInfo>()
        select string.Format ("{0} is {1}", ui.Name, ui.Age);
    }
  }
  #endregion Trying out these toys

  #region Utilities
  /// <summary>
  ///   Extensions to IDictionary.
  /// </summary>
  static class IDictionaryOps
  {
    /// <summary>
    ///   Convert a dictionary to the obvious, if partial, function.
    /// </summary>
    /// <param name='self'>
    ///   The dictionary.
    /// </param>
    /// <returns>
    ///   A function that, given a key as argument, either returns the
    ///   associated value in self, or throws a KeyNotFoundException.
    /// </returns>
    public static Func<K, V> GetFunc<K, V>(this IDictionary<K, V> self)
    {
      return key => self[key];
    }
  }

  /// <summary>
  ///   No data.
  /// </summary>
  public class Unit
  {
    public static readonly Unit Instance = new Unit ();

    private Unit () {}

    public override string ToString ()
    {
      return "(Unit)()";
    }
  }
  #endregion Utilities
}
