//  Copyright (C) 2012  Stephen Compall
// 
//    Licensed under the Apache License, Version 2.0 (the "License");
//    you may not use this file except in compliance with the License.
//    You may obtain a copy of the License at
// 
//        http://www.apache.org/licenses/LICENSE-2.0
// 
//    Unless required by applicable law or agreed to in writing, software
//    distributed under the License is distributed on an "AS IS" BASIS,
//    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//    See the License for the specific language governing permissions and
//    limitations under the License.
using System;
using System.Collections.Generic;
using NUnit.Framework;
using Nocandy.LinqResources.Mystery;
using Nocandy.LinqResources.ReaderV;

namespace Nocandy.LinqResources.ReaderV.Test
{
  /// <summary>
  /// Silly test class to represent user info.
  /// </summary>
  class UserInfo
  {
    public string Name {get; private set;}
    public int Age {get; private set;}
    public UserInfo(string name, int age)
    {
      Name = name;
      Age = age;
    }
  }

  [TestFixture()]
  class TestReaderV
  {
    private static Mystery<Exception, T> Run<T>(ReaderV<UserInfo, T> it)
    {
      return it.Run(new UserInfo(name: "John Doe", age: 42));
    }

    private static ReaderV<UserInfo, String> Describe()
    {
      return from ui in ReaderVOps.Get<UserInfo>()
        select string.Format("{0}, age {1}", ui.Name, ui.Age);
    }

    private static Mystery<Exception, T> MSuccess<T>(T it)
    {
      return AMystery<Exception, T>.Point (it);
    }

    [Test()]
    public void SimpleAccess ()
    {
      Assert.AreEqual(MSuccess ("John Doe, age 42"), Run (Describe()));
    }

    [Test()]
    public void Mapping()
    {
      Assert.AreEqual (MSuccess ("Hello, my name is John Doe, age 42"),
                       Run(from d in Describe () select "Hello, my name is " + d));
    }

    private ReaderV<UserInfo, string> Name()
    {
      return from ui in ReaderVOps.Get<UserInfo>() select ui.Name;
    }

    [Test()]
    public void Sequencing()
    {
      Assert.AreEqual (MSuccess (new {desc = "John Doe, age 42", nm = "John Doe"}),
                       Run(from desc in Describe () from nm in Name ()
                           select new {desc, nm}));
      Assert.AreEqual(new UnknownError<Exception, string>("yup, stopped here"),
                      Run(from desc in Describe ()
                          from oops in "yup, stopped here".RVUnknownError<UserInfo, string>()
                          from nm in Name ()
                          select desc + nm));
    }

    [Test()]
    public void Filtering()
    {
      Assert.AreEqual(MSuccess ("John Doe, age 42"),
                      Run (from desc in Describe ()
                           where desc.StartsWith ("John Doe")
                           select desc));
      Assert.AreEqual("filtered out",
                      Run(from desc in Describe ()
                          where desc.StartsWith ("42")
                          select desc).Fold(e => e, e => "nope", a => "nope"));
    }

    [Test()]
    public void Distributing ()
    {
      Assert.AreEqual(new List<string>{"John Doe, age 42", "John Doe",
                                       "a third entry"},
        Run(new List<ReaderV<UserInfo, string>>{Describe (), Name (),
              "a third entry".RVSuccess<UserInfo, string> ()}.Distribute ())
        .Fold(s => new List<string>{s}, e => new List<string>{e.ToString()},
              a => a));
      Assert.AreEqual(new UnknownError<Exception, IEnumerable<string>>("oops 1"),
                      Run(new List<ReaderV<UserInfo, string>>{
                            Describe(),
                            "oops 1".RVUnknownError<UserInfo, string>(),
                            "oops 2".RVUnknownError<UserInfo, string>()
                          }.Distribute()));
    }

    [Test()]
    public void CategoryIdentity ()
    {
      var readers = new List<ReaderV<UserInfo, string>> {
        Describe (),
        new Exception("bad").RVFailure<UserInfo, string>(),
        "oops".RVUnknownError<UserInfo, string>()
      };
      foreach (var r in readers) {
        Assert.AreEqual (Run (r),
                         Run(r.Compose(ReaderVOps.Identity<string>())));
        Assert.AreEqual (Run (r),
                         Run(ReaderVOps.Identity<UserInfo>().Compose(r)));
      }
    }

    [Test()]
    public void Contramapping ()
    {
      Assert.AreEqual(MSuccess (42),
                      Run (ReaderVOps.Get<int>()
                           .Contramap((UserInfo ui) => ui.Age)));
    }
  }
}

